/*
 * Rjb - Ruby <-> Java Bridge
 * Copyright(c) 2004 arton
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * $Id:$
 * $Log:$
 */

#include <stdlib.h>
#include "ruby.h"
#include "util.h"
#include "jni.h"

#define JVM_TYPE "client"

#if defined(_WIN32) || defined(__CYGWIN__)
 #define JVMDLL "jvm.dll"
 #if defined(__CYGWIN__)
  #define JVMPATH "%s/jre/bin/%s"
  #define DIRSEPARATOR '/'
  #define LOAD_ENVNAME "LD_LIBRARY_PATH"
 #else
  #define JVMPATH "%s\\jre\\bin\\%s"
  #define DIRSEPARATOR '\\'
  #define LOAD_ENVNAME "PATH"
 #endif
#else /* defined(_WIN32) || defined(__CYGWIN__) */
 #ifdef _LP64
  #ifdef ia64
   #define ARCH "ia64"
  #else
   #define ARCH "sparcv9"
  #endif
 #endif
 #if defined(i586) || defined(__i386__)
  #define ARCH "i386"
 #endif
 #ifdef __sparc
  #include <sys/systeminfo.h>
 #endif
 #ifndef ARCH
  #include <sys/systeminfo.h>
 #endif
 #define JVMDLL "libjvm.so"
 #define JVMPATH "%s/jre/lib/%s/%s"
 #define DIRSEPARATOR '/'
 #define LOAD_ENVNAME "LD_LIBRARY_PATH"
#endif

typedef void (*GETDEFAULTJAVAVMINITARGS)(void*);
typedef int (*CREATEJAVAVM)(JavaVM**, void**, void*);

extern JavaVM* jvm;
extern JNIEnv* jenv;
static VALUE jvmdll = Qnil;
static VALUE GetDefaultJavaVMInitArgs;
static VALUE CreateJavaVM;

/*
 * not completed, only valid under some circumstances.
 */
static VALUE setup_path(char* jvmtype)
{
    char* libpath;
    char* ldpath;
    char* orgpath;    
    char* java_home;
    char* jh = getenv("JAVA_HOME");
    if (!jh)
    {
	return Qfalse;
    }
    java_home = ALLOCA_N(char, strlen(jh) + 1);
    strcpy(java_home, jh);
    if (*(java_home + strlen(jh) - 1) == DIRSEPARATOR)
    {
	*(java_home + strlen(jh) - 1) = '\0';
    }
#if defined(_WIN32) || defined(__CYGWIN__)
    libpath = ALLOCA_N(char, strlen(java_home) + sizeof(JVMPATH)
		       + strlen(jvmtype) + 1);
    sprintf(libpath, JVMPATH, java_home, jvmtype);
#else /* not Windows */
    libpath = ALLOCA_N(char, strlen(java_home) + sizeof(JVMPATH)
		       + strlen(ARCH) + strlen(jvmtype) + 1);
    sprintf(libpath, JVMPATH, java_home, ARCH, jvmtype);
#endif
    orgpath = getenv(LOAD_ENVNAME);
    if (!orgpath)
    {
	ldpath = libpath;
    }
    else
    {
	ldpath = ALLOCA_N(char, strlen(orgpath) + strlen(libpath) + 2);
	sprintf(ldpath, "%s%s%s", orgpath, PATH_SEP, libpath);
    }
    ruby_setenv(LOAD_ENVNAME, ldpath);
    return Qtrue;
}

static void load_jvm()
{
    VALUE dl;
    rb_require("dl");
    dl = rb_eval_string("DL");
    jvmdll = rb_funcall(dl, rb_intern("dlopen"),
			1, rb_str_new(JVMDLL, sizeof(JVMDLL)));
    GetDefaultJavaVMInitArgs = rb_funcall(jvmdll, rb_intern("sym"),
			  2, rb_str_new2("JNI_GetDefaultJavaVMInitArgs"),
			     rb_str_new2("0p"));
    CreateJavaVM = rb_funcall(jvmdll, rb_intern("sym"),
			      2, rb_str_new2("JNI_CreateJavaVM"),
			         rb_str_new2("pppp"));
}

int create_jvm(JDK1_1InitArgs* vm_args, char* userpath)
{
    VALUE ptr;
    int len;
    GETDEFAULTJAVAVMINITARGS initargs;
    CREATEJAVAVM createjavavm;
    if (jvmdll == Qnil)
    {
	if (setup_path(JVM_TYPE) == Qfalse)
	{
	    return -1;
	}

	load_jvm();
    }

#if defined(_WIN32) || defined(__CYGWIN__)
    ptr = rb_funcall(GetDefaultJavaVMInitArgs, rb_intern("to_i"), 0);
    initargs = *(GETDEFAULTJAVAVMINITARGS*)FIX2INT(ptr);
    initargs(vm_args);
#endif

    len = strlen(userpath);
    if (vm_args->classpath)
    {
	char* newpath = ALLOCA_N(char, len + strlen(vm_args->classpath) + 2);
	sprintf(newpath, "%s%c%s", userpath, PATH_SEP, vm_args->classpath);
	vm_args->classpath = newpath;
    }
    else
    {
	vm_args->classpath = userpath;
    }
    ptr = rb_funcall(CreateJavaVM, rb_intern("to_i"), 0);
    createjavavm = *(CREATEJAVAVM*)FIX2INT(ptr);
    return createjavavm(&jvm, (void**)&jenv, vm_args);
}

